/*  Copyright (C) 1990, Jim Crammond, Imperial College. All rights reserved.  */

/*
 *  which data area/processor tests
 */
#ifdef PARALLEL
#define	ismaster(pr)	(pr == Pr0)
#else
#define	ismaster(pr)	(1)		/* always true */
#endif PARALLEL


/*
 *  dereference, wait for argument
 */
#define	deref(x)				\
	while ( IsRef(x) && !IsUnb(*ToPtr(x)) )	\
		x = (Word) *ToPtr(x);


#define wait_for_argument(x)			\
	while (IsRef(x))			\
	{	if (IsUnb(*ToPtr(x)))		\
		{	VarTbl[Nsuspv++] = x;	\
			return(SUSPEND);	\
		}				\
		x = (Word) *ToPtr(x);		\
	}


/*
 *  non-local variable assignment:  must lock to set variable
 *  (we use atomic exchange facility if possible)  and must
 *  wake any processes suspended on that variable.
 */
#ifdef PARALLEL

#define assign(ref, val)			\
	{	register Word	oval;		\
						\
		Exchange(ref, val, oval);	\
						\
		if (!IsUnb(oval))		\
		{	if (!unify(oval,val))	\
				return_fail;	\
		}				\
		else if (oval != AsUnb(0))	\
			wake(UnbVal(oval));	\
	}


#define	return_fail	return(FAIL)

#if defined(symmetry) || defined(butterfly)
#define Exchange(ref, val, oval)	oval = exchange(ref, val)

#else
	/*  simulate an atomic exchange  */

#define Exchange(ref, val, oval)		\
	{	register char	*vlck;		\
						\
		vlck = ptrtolck(ref);		\
		lock(vlck);			\
		oval = *ToPtr(ref);		\
		*ToPtr(ref) = val;		\
		unlock(vlck);			\
	}
#endif


#else !PARALLEL

#define assign(ref, val)			\
	{	Process	*ps_list;		\
						\
		ps_list = UnbVal(*ToPtr(ref));	\
		*ToPtr(ref) = val;		\
						\
		if (ps_list)			\
			wake(ps_list);		\
	}

#endif PARALLEL


/*
 *  add a process to the local run queue
 */
#ifdef PRIVATE_QUEUE

#define enqueue_process(ps)			\
	*q_back++ = ps;				\
	if (q_back == PR->q_end)		\
		q_back = PR->q_bot;		\
	if (PR->q_back == PR->q_front)		\
	{	/*  make a process public  */	\
		PR->q_back++;			\
		if (PR->q_back == PR->q_end)	\
			PR->q_back = PR->q_bot;	\
	}

#else

#define enqueue_process(ps)			\
	*PR->q_back = ps;			\
	PR->q_back++;				\
	if (PR->q_back == PR->q_end)		\
		PR->q_back = PR->q_bot;		\

#endif PRIVATE_QUEUE


/*
 *  argument stack allocation
 */
#define alloc_stack(i)				\
	m_st->m_top;				\
	m_st->m_top += i


#define dealloc_stack(p, n)			\
	if ((p + n) == SP)			\
	{	SP = p;				\
		while (*--SP == 0)		\
			;			\
		SP++;				\
	}					\
	else					\
	{	register Word *sp = p;		\
		register int  sn  = n;		\
						\
		while (sn-- > 0)		\
			*sp++ = 0;		\
	}


/*
 *  process stack allocation
 */
#define	alloc_ps(ps)				\
	if (p_free)				\
	{	ps = p_free;			\
		p_free = p_free->link;		\
	}					\
	else					\
	{	ps = p_top++;			\
		if (p_top >= p_end)		\
			alloc_proc();		\
	}

#define dealloc_ps(ps)				\
	{	ps->link = p_free;		\
		p_free = ps;			\
	}


/*
 *  block argument copying
 */
#define copyargs(from, to, len)		\
	{	register Word *p, *q;	\
		register int n;		\
					\
		p = from;		\
		q = to;			\
		n = len;		\
		while (n-- > 0)		\
			*q++ = *p++;	\
	}

#define pushargs(from, len)		\
	{	register Word *p;	\
		register int n;		\
					\
		p = from;		\
		n = len;		\
		while (n-- > 0)		\
			*SP++ = *p++;	\
	}
